"use strict";
/**
 * @copyright Cube Dev, Inc.
 * @license Apache-2.0
 * @fileoverview The `AthenaDriver` and related types declaration.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.AthenaDriver = void 0;
const shared_1 = require("@cubejs-backend/shared");
const client_athena_1 = require("@aws-sdk/client-athena");
const stream = __importStar(require("stream"));
const base_driver_1 = require("@cubejs-backend/base-driver");
const SqlString = __importStar(require("sqlstring"));
const url_1 = require("url");
function applyParams(query, params) {
    return SqlString.format(query, params);
}
class AthenaDriver extends base_driver_1.BaseDriver {
    /**
     * Returns default concurrency value.
     */
    static getDefaultConcurrency() {
        return 10;
    }
    config;
    athena;
    schema;
    /**
     * Class constructor.
     */
    constructor(config = {}) {
        super({
            testConnectionTimeout: config.testConnectionTimeout,
        });
        const dataSource = config.dataSource ||
            (0, shared_1.assertDataSource)('default');
        const accessKeyId = config.accessKeyId ||
            (0, shared_1.getEnv)('athenaAwsKey', { dataSource });
        const secretAccessKey = config.secretAccessKey ||
            (0, shared_1.getEnv)('athenaAwsSecret', { dataSource });
        const { schema, ...restConfig } = config;
        this.schema = schema ||
            (0, shared_1.getEnv)('dbName', { dataSource }) ||
            (0, shared_1.getEnv)('dbSchema', { dataSource });
        this.config = {
            credentials: accessKeyId && secretAccessKey
                ? { accessKeyId, secretAccessKey }
                : undefined,
            ...restConfig,
            region: config.region ||
                (0, shared_1.getEnv)('athenaAwsRegion', { dataSource }),
            S3OutputLocation: config.S3OutputLocation ||
                (0, shared_1.getEnv)('athenaAwsS3OutputLocation', { dataSource }),
            workGroup: config.workGroup ||
                (0, shared_1.getEnv)('athenaAwsWorkgroup', { dataSource }) ||
                'primary',
            catalog: config.catalog ||
                (0, shared_1.getEnv)('athenaAwsCatalog', { dataSource }),
            database: config.database ||
                (0, shared_1.getEnv)('dbName', { dataSource }),
            exportBucket: config.exportBucket ||
                (0, shared_1.getEnv)('dbExportBucket', { dataSource }),
            pollTimeout: (config.pollTimeout ||
                (0, shared_1.getEnv)('dbPollTimeout', { dataSource }) ||
                (0, shared_1.getEnv)('dbQueryTimeout', { dataSource })) * 1000,
            pollMaxInterval: (config.pollMaxInterval ||
                (0, shared_1.getEnv)('dbPollMaxInterval', { dataSource })) * 1000,
            exportBucketCsvEscapeSymbol: (0, shared_1.getEnv)('dbExportBucketCsvEscapeSymbol', { dataSource }),
        };
        if (this.config.exportBucket) {
            this.config.exportBucket =
                AthenaDriver.normalizeS3Path(this.config.exportBucket);
        }
        if (typeof this.config.readOnly === 'undefined') {
            // If Export bucket configuration is in place we want to always use it instead of batching
            this.config.readOnly = !this.isUnloadSupported();
        }
        this.athena = new client_athena_1.Athena(this.config);
    }
    /**
     * Driver read-only flag.
     */
    readOnly() {
        return !!this.config.readOnly;
    }
    /**
     * Returns driver's capabilities object.
     */
    capabilities() {
        return {
            unloadWithoutTempTable: true,
            incrementalSchemaLoading: true,
        };
    }
    /**
     * Test driver's connection.
     */
    async testConnection() {
        await this.athena.getWorkGroup({
            WorkGroup: this.config.workGroup
        });
    }
    /**
     * Executes a query and returns either query result memory data or
     * query result stream, depending on options.
     */
    async downloadQueryResults(query, values, options) {
        if (!options.streamImport) {
            return this.memory(query, values);
        }
        else {
            return this.stream(query, values, options);
        }
    }
    /**
     * Executes query and returns table memory data that includes rows
     * and queried fields types.
     */
    async memory(query, values) {
        const qid = await this.startQuery(query, values);
        await this.waitForSuccess(qid);
        const iter = this.lazyRowIterator(qid, query, true);
        const types = ((await iter.next()).value);
        const rows = [];
        for await (const row of iter) {
            rows.push(row);
        }
        return { types, rows };
    }
    /**
     * Returns stream table object that includes query result stream and
     * queried fields types.
     */
    async stream(query, values, options) {
        const qid = await this.startQuery(query, values);
        await this.waitForSuccess(qid);
        const iter = this.lazyRowIterator(qid, query, true);
        const types = ((await iter.next()).value);
        return {
            rowStream: stream.Readable.from(iter, {
                highWaterMark: options.highWaterMark,
            }),
            types,
            release: async () => { },
        };
    }
    /**
     * Executes query and rerutns queried rows.
     */
    async query(query, values, _options) {
        const qid = await this.startQuery(query, values);
        await this.waitForSuccess(qid);
        const rows = [];
        for await (const row of this.lazyRowIterator(qid, query)) {
            rows.push(row);
        }
        return rows;
    }
    /**
     * Executes query and returns async generator that yields queried
     * rows.
     */
    async *lazyRowIterator(qid, query, withTypes) {
        let isFirstBatch = true;
        let columnInfo = [];
        for (let results = await this.athena.getQueryResults(qid); results; results = results.NextToken
            ? (await this.athena.getQueryResults({ ...qid, NextToken: results.NextToken }))
            : undefined) {
            let rows = results.ResultSet?.Rows ?? [];
            if (isFirstBatch) {
                if (withTypes) {
                    yield this.mapTypes(results.ResultSet?.ResultSetMetadata?.ColumnInfo);
                }
                isFirstBatch = false;
                // Athena returns the columns names in first row, skip it.
                rows = rows.slice(1);
                columnInfo = /SHOW COLUMNS/.test(query) // Fix for getColumns method
                    ? [{ Name: 'column' }]
                    : (0, shared_1.checkNonNullable)('ColumnInfo', results.ResultSet?.ResultSetMetadata?.ColumnInfo).map(info => ({ Name: (0, shared_1.checkNonNullable)('Name', info.Name) }));
            }
            for (const row of rows) {
                const fields = {};
                columnInfo
                    .forEach((c, j) => {
                    const r = row.Data;
                    fields[c.Name] = (r === null ||
                        r === undefined ||
                        r[j].VarCharValue === undefined) ? null : r[j].VarCharValue;
                });
                yield fields;
            }
        }
    }
    /**
     * Save pre-aggregation data into a temp table.
     */
    async loadPreAggregationIntoTable(preAggregationTableName, loadSql, params) {
        if (this.config.S3OutputLocation === undefined) {
            throw new Error('Unload is not configured. Please define CUBEJS_AWS_S3_OUTPUT_LOCATION env var ');
        }
        const qid = await this.startQuery(loadSql, params);
        await this.waitForSuccess(qid);
    }
    /**
     * Determines whether export bucket feature is configured or not.
     */
    async isUnloadSupported() {
        return this.config.exportBucket !== undefined;
    }
    /**
     * Returns to the Cubestore an object with links to unloaded to the
     * export bucket data.
     */
    async unload(tableName, options) {
        if (!this.config.exportBucket) {
            throw new Error('Export bucket is not configured.');
        }
        const types = options.query
            ? await this.unloadWithSql(tableName, options)
            : await this.unloadWithTable(tableName);
        const csvFile = await this.getCsvFiles(tableName);
        return {
            exportBucketCsvEscapeSymbol: this.config.exportBucketCsvEscapeSymbol,
            csvFile,
            types,
            csvNoHeader: true,
            csvDelimiter: '^A',
        };
    }
    /**
     * Unload data from a SQL query to an export bucket.
     */
    async unloadWithSql(tableName, unloadOptions) {
        const columns = await this.queryColumnTypes(unloadOptions.query.sql, unloadOptions.query.params);
        const unloadSql = `
      UNLOAD (${unloadOptions.query.sql})
      TO '${this.config.exportBucket}/${tableName}'
      WITH (
        format = 'TEXTFILE',
        compression='GZIP'
      )`;
        const qid = await this.startQuery(unloadSql, unloadOptions.query.params);
        await this.waitForSuccess(qid);
        await this.athena.getQueryResults(qid);
        return columns;
    }
    /**
     * Unload data from a temp table to an export bucket.
     */
    async unloadWithTable(tableName) {
        const types = await this.tableColumnTypes(tableName);
        const columns = types.map(t => t.name).join(', ');
        const unloadSql = `
      UNLOAD (SELECT ${columns} FROM ${tableName})
      TO '${this.config.exportBucket}/${tableName}'
      WITH (
        format = 'TEXTFILE',
        compression='GZIP'
      )`;
        const qid = await this.startQuery(unloadSql, []);
        await this.waitForSuccess(qid);
        return types;
    }
    /**
     * Returns an array of queried fields meta info.
     */
    async queryColumnTypes(sql, params) {
        const unloadSql = `${sql} LIMIT 0`;
        const qid = await this.startQuery(unloadSql, params || []);
        await this.waitForSuccess(qid);
        const results = await this.athena.getQueryResults(qid);
        const columns = this.mapTypes(results.ResultSet?.ResultSetMetadata?.ColumnInfo);
        return columns;
    }
    /**
     * Converts Athena to generic types and returns an array of queried
     * fields meta info.
     */
    mapTypes(fields) {
        return fields.map((field) => ({ name: field.Name, type: this.toGenericType(field.Type || 'text') }));
    }
    /**
     * Returns an array of signed URLs of the unloaded csv files.
     */
    async getCsvFiles(tableName) {
        const { bucket, prefix } = AthenaDriver.splitS3Path(`${this.config.exportBucket}/${tableName}`);
        return this.extractUnloadedFilesFromS3({
            credentials: this.config.credentials,
            region: this.config.region,
        }, bucket, prefix.slice(1));
    }
    informationSchemaQuery() {
        if (this.schema) {
            return `${super.informationSchemaQuery()} AND columns.table_schema = '${this.schema}'`;
        }
        return super.informationSchemaQuery();
    }
    async tablesSchema() {
        const tablesSchema = await super.tablesSchema();
        const viewsSchema = await this.viewsSchema(tablesSchema);
        return this.mergeSchemas([tablesSchema, viewsSchema]);
    }
    async startQuery(query, values) {
        const queryString = applyParams(query, (values || []).map(s => (typeof s === 'string' ? {
            toSqlString: () => SqlString.escape(s).replace(/\\\\([_%])/g, '\\$1').replace(/\\'/g, '\'\'')
        } : s)));
        const request = {
            QueryString: queryString,
            WorkGroup: this.config.workGroup,
            ResultConfiguration: {
                OutputLocation: this.config.S3OutputLocation
            },
            ...(this.config.catalog || this.config.database ? {
                QueryExecutionContext: {
                    Catalog: this.config.catalog,
                    Database: this.config.database
                }
            } : {})
        };
        const { QueryExecutionId } = await this.athena.startQueryExecution(request);
        return { QueryExecutionId: (0, shared_1.checkNonNullable)('StartQueryExecution', QueryExecutionId) };
    }
    async checkStatus(qid) {
        const queryExecution = await this.athena.getQueryExecution(qid);
        const status = queryExecution.QueryExecution?.Status?.State;
        if (status === 'FAILED') {
            throw new Error(queryExecution.QueryExecution?.Status?.StateChangeReason);
        }
        if (status === 'CANCELLED') {
            throw new Error('Query has been cancelled');
        }
        return status === 'SUCCEEDED';
    }
    async waitForSuccess(qid) {
        const startedTime = Date.now();
        for (let i = 0; Date.now() - startedTime <= this.config.pollTimeout; i++) {
            if (await this.checkStatus(qid)) {
                return;
            }
            await (0, shared_1.pausePromise)(Math.min(this.config.pollMaxInterval, 500 * i));
        }
        throw new Error(`Athena job timeout reached ${this.config.pollTimeout}ms`);
    }
    async viewsSchema(tablesSchema) {
        const isView = (table) => !tablesSchema[table.schema]
            || !tablesSchema[table.schema][table.name];
        const allTables = await this.getAllTables();
        const arrViewsSchema = await Promise.all(allTables
            .filter(isView)
            .map(table => this.getColumns(table)));
        return this.mergeSchemas(arrViewsSchema);
    }
    async getAllTables() {
        let allTablesQuery = `
      SELECT table_schema AS schema, table_name AS name
      FROM information_schema.tables
      WHERE tables.table_schema NOT IN ('information_schema', 'mysql', 'performance_schema', 'sys')
    `;
        if (this.schema) {
            allTablesQuery = `${allTablesQuery} AND tables.table_schema = '${this.schema}'`;
        }
        const rows = await this.query(allTablesQuery, []);
        return rows;
    }
    async getColumns(table) {
        const data = await this.query(`SHOW COLUMNS IN \`${table.schema}\`.\`${table.name}\``, []);
        return {
            [table.schema]: {
                [table.name]: data.map(({ column }) => {
                    const [name, type] = column.split('\t');
                    return { name, type, attributes: [] };
                })
            }
        };
    }
    mergeSchemas(arrSchemas) {
        const result = {};
        arrSchemas.forEach(schemas => {
            Object.keys(schemas).forEach(schema => {
                Object.keys(schemas[schema]).forEach((name) => {
                    if (!result[schema])
                        result[schema] = {};
                    if (!result[schema][name])
                        result[schema][name] = schemas[schema][name];
                });
            });
        });
        return result;
    }
    static normalizeS3Path(path) {
        // Remove trailing /
        path = path.replace(/\/+$/, '');
        // Prepend s3:// prefix to plain bucket
        if (!path.startsWith('s3://')) {
            return `s3://${path}`;
        }
        return path;
    }
    static splitS3Path(path) {
        const url = new url_1.URL(path);
        return {
            bucket: url.host,
            prefix: url.pathname
        };
    }
}
exports.AthenaDriver = AthenaDriver;
//# sourceMappingURL=AthenaDriver.js.map